from rulexai import rule
import unittest
import rulekit
from rulekit.main import RuleKit
from rulexai.explainer import RuleExplainer
from sklearn.datasets import load_iris
import pandas as pd
from rulekit.classification import RuleClassifier
from rulekit.params import Measures
from sklearn.tree import DecisionTreeClassifier
from Orange.data.table import Table
from Orange.data.pandas_compat import table_to_frame
from Orange.classification import CN2UnorderedLearner

import os 
from pathlib import Path
import numpy as np

class TestFunctionalities(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        RuleKit.init()

    def test_creating_model_with_rules_RuleKit(self):
        data = load_iris()
        x = pd.DataFrame(data.data, columns=data.feature_names)
        y = pd.DataFrame(data.target.astype(str), columns=["label"]) 

        clf = RuleClassifier(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
        )
        clf.fit(x, data.target.astype(str))

        explainer = RuleExplainer(clf, x, y, type="classification")

        assert len(clf.model.rules) == len(explainer.model.rules)

    
    def test_creating_model_with_rules_Sklearn(self):
        clf = DecisionTreeClassifier(random_state=0)
        data = load_iris()
        clf.fit(data.data, data.target)

        x = pd.DataFrame(data.data, columns=data.feature_names)
        y = pd.DataFrame(data.target.astype(str), columns=["class"]) 

        explainer = RuleExplainer(clf, x, y, type="classification")
        
        assert len(explainer.model.rules) > 0 

    def test_creating_model_with_rules_Orange(self):
        data = Table("iris")
        learner = CN2UnorderedLearner()
        clf = learner(data)
        df= table_to_frame(data)
        x = df.drop(['iris'], axis=1)
        y = df['iris']

        explainer = RuleExplainer(clf, x, y, type="classification")
        
        assert len(explainer.model.rules) > 0 


    def test_features_importances(self):

        results_path = "resources/features_importances.csv"
        results_path = os.path.join(Path(__file__).resolve().parent, results_path)

        data = load_iris()
        x = pd.DataFrame(data.data, columns=data.feature_names)
        y = pd.DataFrame(data.target.astype(str), columns=["label"]) 

        clf = RuleClassifier(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
        )
        clf.fit(x, data.target.astype(str))

        explainer = RuleExplainer(clf, x, y, type="classification")
        explainer.explain()

        features_importances = explainer.feature_importances_
        features_importances.replace("-", np.nan, inplace = True)
        features_importances = features_importances.round(4)


        features_importances_ground_truth = pd.read_csv(results_path, sep = ";")
        features_importances_ground_truth.replace("-", np.nan, inplace = True)
        features_importances_ground_truth = features_importances_ground_truth.round(4)

        assert features_importances.equals(features_importances_ground_truth)


    def test_prepare_binary_dataset(self):
        data = load_iris()
        x = pd.DataFrame(data.data, columns=data.feature_names)
        y = pd.DataFrame(data.target.astype(str), columns=["label"]) 

        clf = RuleClassifier(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
        )
        clf.fit(x, data.target.astype(str))

        explainer = RuleExplainer(clf, x, y, type="classification")
        explainer.explain()

        binary_dataset = explainer.fit_transform(x)
        binary_dataset_10 = explainer.transform(x.iloc[0:10])

 
        binary_dataset_TOP_50 = explainer.fit_transform(x, selector=0.5)

        assert binary_dataset.shape[0] == x.shape[0]
        assert binary_dataset_10.shape[0] == 10
        assert binary_dataset.shape[1] == 14
        assert binary_dataset_TOP_50.shape[0] == x.shape[0]
        assert binary_dataset_TOP_50.shape[1] == 7



    def test_get_rules_covering_example_and_get_rules(self):
        data = load_iris()
        x = pd.DataFrame(data.data, columns=data.feature_names)
        y = pd.DataFrame(data.target.astype(str), columns=["label"]) 

        clf = RuleClassifier(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
        )
        clf.fit(x, data.target.astype(str))

        explainer = RuleExplainer(clf, x, y, type="classification")

        rules_covering_example_0 = explainer.get_rules_covering_example(x.iloc[0,:], y.iloc[0,:])
        rules_covering_example_100 = explainer.get_rules_covering_example(x.iloc[100,:], y.iloc[100,:])

        assert len(rules_covering_example_0) == 1
        assert len(rules_covering_example_100) == 2
        assert len(explainer.get_rules()) == len(explainer.get_rules_with_basic_conditions())
        assert len(explainer.get_rules()) == len(clf.model.rules)


if __name__ == '__main__':
    unittest.main()


from .importances import (
    ClassificationConditionImportance,
    RegressionConditionImportance,
    SurvivalConditionImportance,
    ConditionImportance,
)
from .models import ClassificationModel, RegressionModel, SurvivalModel, BlackBoxModel
import pandas as pd
import numpy as np
from typing import Union, List
import matplotlib.pyplot as plt

from .reduct import Reduct

Labels = Union[pd.DataFrame, pd.Series]



class BaseExplainer:
    """:meta private:"""
    def __init__(
        self, model, X: pd.DataFrame, y: Labels, type: str = "classification"
    ) -> None:


        self.model = model
        self.X = X
        self.y = y
        self.type = type

        self.condition_importances_ = None
        self.feature_importances_ = None
        self.if_basic_conditions = None

        self.condition_importance_class = None                            
        self._conditions_importances_for_training_set = None 
        self._basic_conditions_importances_for_training_set = None


    def explain(self, measure: str = "C2", basic_conditions: bool = False):
        """Compute conditions importances. The importances of a conditions are computed base on: \n
        Marek Sikora: Redefinition of Decision Rules Based on the Importance of Elementary Conditions Evaluation. Fundam. Informaticae 123(2): 171-197 (2013) \n
        https://dblp.org/rec/journals/fuin/Sikora13.html

        Parameters
        ----------
        measure: str
            Specifies the measure that is used to evaluate the quality of the rules. Possible measures for classification and regression problem are: C2, Lift, Correlation. Default: C2. It is not possible to select a measure for the survival problem, the LogRank test is used by default 
        basic_conditions : bool
            Specifies whether to evaluate the conditions contained in the input rules, or to break the conditions in the rules into base conditions so that individual conditions do not overlap
        Returns
        -------
        self: Explainer
            Fitted explainer with calculated conditions

        """
        self.if_basic_conditions = basic_conditions
        self.condition_importances_ = self._determine_condition_importances(measure)
        self.feature_importances_ = self._determine_feature_importances(self.condition_importances_)

        return self

    def fit_transform(
        self, X: pd.DataFrame, selector=None, y=None, POS=None) -> pd.DataFrame:

        """Creates a dataset based on given dataset in which the examples, instead of being described by the original attributes, will be described with the specified conditions - it will be a set with binary attributes determining whether a given example meets a given condition. It can be considered as kind of dummification.
        Thanks to this function you can discretize data and get rid of missing values. It can be used as prestep for others algorithms.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples from which you want to create binary dataset. Should have the same columns and columns order as X specified when creating Explainer
        selector : string/float
            Specifies on what basis to select the conditions from the rules that will be included as attributes in the transformed set.   
            If None all conditions will be included in the transformed set. If number 0-1 percent of the most important conditions will be selected based on condition importance ranking. If "reduct" the reduct of the conditions set will be selected. Preferably, the option with the percentage of most important conditions will be selected.
        y: Union[pd.DataFrame, pd.Series]
            Only if selector = "reduct".The target values for input sample, used in the determination of the reduct
        POS: float
            Only if selector = "reduct".Target reduct POS
        Returns
        -------
        X_transformed: pd.DataFrame
            Transformed dataset

        """
        
        helper = ConditionImportance(self.model.rules, X, None, self.if_basic_conditions)
        if self.if_basic_conditions:
            rules = helper.split_conditions_in_rules(self.model.rules)
            conditions = helper._get_conditions_with_rules(rules)
        else:
            conditions = helper._get_conditions_with_rules(
                self.model.rules
            )

        binary_dataset = self._prepare_binary_dataset(X, conditions)

        if selector=="reduct":
            reduct = Reduct()
            binary_dataset = reduct.get_reduct(binary_dataset,y,POS)
        elif not selector is None:
            binary_dataset = self._get_top_conditions(binary_dataset, selector)
        
        chosen_conditions_names = binary_dataset.columns
        self.conditions_names = chosen_conditions_names
        self.conditions = []
        for condition in conditions:
            if str(condition) in chosen_conditions_names:
                self.conditions.append(condition)

        return binary_dataset

    def transform(
        self, X: pd.DataFrame) -> pd.DataFrame:

        """Creates a dataset based on given dataset in which the examples, instead of being described by the original attributes, will be described with the specified conditions - it will be a set with binary attributes determining whether a given example meets a given condition. It can be considered as kind of dummification.
        Thanks to this function you can discretize data and get rid of missing values. It can be used as prestep for others algorithms.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples from which you want to create binary dataset. Should have the same columns and columns order as X given in fit_transform
        Returns
        -------
        X_transformed: pd.DataFrame
            Transformed dataset

        """
        transformed_dataset = self._prepare_binary_dataset(X, self.conditions)
            
        return transformed_dataset[self.conditions_names]

   
    def get_rules_covering_example(self, x: pd.DataFrame, y: Labels) -> List[str]:
        """Return rules that covers the given example

        Parameters
        ----------
        x : pd.DataFrame
            The input sample.
        y : Union[pd.DataFrame, pd.Series]
            The target values for input sample.
        Returns
        -------
        rules: List[str]
            Rules that covers the given example

        """
        rules_covering_example = []
        for rule in self.model.rules:
            if rule.premise.evaluate(x):
                rules_covering_example.append(rule)

        return rules_covering_example


    def local_explainability(self, x: pd.DataFrame, y: Labels, plot: bool = False):
        """Displays information about the local explanation of the example: the rules that cover the given example and the importance of the conditions contained in these rules
        
        Parameters
        ----------
        x : pd.DataFrame
            The input sample.
        y : Union[pd.DataFrame, pd.Series]
            The target values for input sample.
        plot : bool
            If True the importance of the conditions will also be shown in the chart. Default: False
        """
        rules_covering_example = self.get_rules_covering_example(x,y)

        print("Example:")
        print(pd.concat([x,y]))
        print("")

        print("Rules that covers this example:")
        for rule in rules_covering_example:
            print(rule)
        print("")
            
        conditions_importances = self.condition_importances_.copy()   
        
        classes_with_conditions = dict()
        for rule in rules_covering_example:
            

            if rule.consequence.left in classes_with_conditions.keys():
                conditions = classes_with_conditions[rule.consequence.left]
            else:
                conditions = []

            conditions.extend(rule.premise.get_subconditions())
            conditions = list(map(str, conditions))
            classes_with_conditions[rule.consequence.left] = conditions



        if self.type == "classification":
            importances_for_covering_rules = pd.DataFrame()
            for j in range(0, conditions_importances.shape[1], 2):
                class_in_consequences = False
                tmp_df = conditions_importances.iloc[:, j : j + 2]
                for cl in classes_with_conditions.keys():
                    if cl in tmp_df.columns[0]:
                        class_in_consequences = True
                        class_name = cl

                if class_in_consequences:
                    tmp_df.loc[~tmp_df[tmp_df.columns[0]].isin(classes_with_conditions[class_name]), tmp_df.columns[0]] = np.NaN
                    tmp_df.dropna(inplace=True)
                    tmp_df.reset_index(drop = True, inplace = True)
                    importances_for_covering_rules = pd.concat([importances_for_covering_rules, tmp_df], ignore_index=False, axis=1)

            importances_for_covering_rules = importances_for_covering_rules.replace(np.nan, "-")
        else:
            conditions_importances.loc[~conditions_importances[conditions_importances.columns[0]].isin(conditions), conditions_importances.columns[0]] = np.NaN
            conditions_importances.dropna(inplace=True)
            importances_for_covering_rules = conditions_importances.reset_index(drop = True)



        print("Importances of the conditions from rules covering the example")
        print(importances_for_covering_rules)

        if plot:
            self.plot_importances(importances_for_covering_rules)

        return importances_for_covering_rules

    def get_rules(self):
        """Return rules from model

        Returns
        -------
        rules: List[str]
            Rules from model
        """
        rules = []
        for rule in self.model.rules:
            rules.append(rule.__str__())
        return rules


    def get_rules_with_basic_conditions(self):
        """Return rules from model with conditions broken down into base conditions so that individual conditions do not overlap

        Returns
        -------
        rules: List[str]
            Rules from the model containing the base conditions 
        """
        rules_with_basic_conditions = []
        helper = ConditionImportance(self.model.rules, self.X, self.y, True)
        rules = helper.rules
        for rule in rules:
            rules_with_basic_conditions.append(rule.__str__())
        return rules_with_basic_conditions


    def plot_importances(self, importances: pd.DataFrame):
        """Plot importances
        Parameters
        ----------
        importances : pd.DataFrame
            Feature/Condition importances to plot.
        """
        
        if "attributes" in importances.columns[0]:
            title = "Feature Importance"
        else:
            title = "Condition Importance"

        plots_number = int(importances.shape[1]/2)

        if self.type == "classification" and plots_number > 1:

            fig, axs = plt.subplots(1,plots_number, sharex = True)
            i = 0
            ticks_number = importances.shape[0]
            y_ticks = np.arange(0, ticks_number)
            for j in range(0, importances.shape[1], 2):

                tmp_df = importances.iloc[:, j : j + 2]
                tmp_df = tmp_df.replace("-", np.nan).dropna()
                tmp_df.sort_values(inplace=True, by=tmp_df.columns[1])
                    
                ticks_values = tmp_df.iloc[:, 1].to_list()
                ticks_all = [0 for _ in range(len(ticks_values),ticks_number)]
                ticks_all.extend(ticks_values)  
                labels = tmp_df.iloc[:, 0].to_list()
                labels_all = ["" for _ in range(len(labels),ticks_number)]
                labels_all.extend(labels)

                colors = ["green" if y >= 0 else "red" for y in ticks_all]

                axs[i].barh(y_ticks, ticks_all, color = colors)
                axs[i].set_yticks(y_ticks)
                axs[i].set_yticklabels(labels_all)
                class_name = tmp_df.columns[0].split(" | ")[0]
                axs[i].set_title(f"Importance for class: {class_name}")
                axs[i].set_xlabel(f"Importance")
                i+=1

            fig.subplots_adjust(wspace = 0.5)
            plt.show()

        else:

            tmp_df = importances.sort_values(by=importances.columns[1])     
            colors = ["green" if y >= 0 else "red" for y in tmp_df.iloc[:,1].to_list()] 
            y_ticks = np.arange(0, len(tmp_df))
            fig, ax = plt.subplots()
            ax.barh(y_ticks, tmp_df.iloc[:, 1], color = colors)
            ax.set_yticks(y_ticks)
            ax.set_yticklabels(tmp_df.iloc[:, 0])
            ax.set_title(f"{title}")
            ax.set_xlabel(f"Importance")

            plt.show()


    
    def _determine_condition_importances(self, measure: str = "C2", X = None, y = None):
        if X is None:
            X = self.X
        if y is None:
            y = self.y

        if self.type == "regression":
            self.condition_importance_class = RegressionConditionImportance(
                rules = self.model.rules, data = X, labels = y, if_split = self.if_basic_conditions, measure = measure
            )
        elif self.type == "survival":
            self.condition_importance_class = SurvivalConditionImportance(
                rules = self.model.rules, data = X, labels = y, if_split = self.if_basic_conditions, measure = measure
            )
        else:
            self.condition_importance_class = ClassificationConditionImportance(
                rules = self.model.rules, data = X, labels = y, if_split = self.if_basic_conditions, measure = measure
            )
        return self.condition_importance_class.condition_importances()

    def _determine_feature_importances(self, conditions_importances):
        feature_importances = pd.DataFrame()

        if self.type == "classification":
            
            for j in range(0, conditions_importances.shape[1], 2):
                class_importances = (
                    conditions_importances.iloc[:, j : j + 2]
                    .replace("-", np.nan)
                    .dropna()
                )
                importances_df_tmp = pd.DataFrame()
                class_importances.iloc[:, 0] = class_importances.iloc[:, 0].apply(
                    lambda x: x.split(" = ")[0]
                )
                class_importances = (
                    class_importances.groupby(class_importances.columns[0])
                    .sum()
                    .reset_index()
                )
                class_importances.sort_values(
                    class_importances.columns[1], ascending=False, inplace=True
                )
                class_importances.reset_index(drop=True, inplace= True)

                class_name, _ = class_importances.columns[0].split(" | ")
                importances_df_tmp[class_name + " | attributes"] = pd.Series(class_importances[
                    class_importances.columns[0]
                ])
                importances_df_tmp[class_name + " | importances"] = pd.Series(class_importances[
                    class_importances.columns[1]
                ])
           
                feature_importances = pd.concat(
                    [feature_importances, importances_df_tmp],
                    ignore_index=False,
                    axis=1
                )
             
            feature_importances = feature_importances.replace(np.nan, "-")
                            
        else:

            importances_df = conditions_importances.copy()
            importances_df.iloc[:, 0] = importances_df.iloc[:, 0].apply(
                lambda x: x.split(" = ")[0]
            )
            importances_df = (
                                       
                importances_df.groupby(importances_df.columns[0]).sum().reset_index()
            )
            importances_df.sort_values(
                importances_df.columns[1], ascending=False, inplace=True
            )

            importances_df.rename(
                columns={
                    importances_df.columns[0]: "attributes",
                    importances_df.columns[1]: "importances",
                },
                inplace=True,
            )

            feature_importances = importances_df.replace(np.nan, "-")

        return feature_importances

    def _prepare_binary_dataset(
        self, X: pd.DataFrame, conditions) -> pd.DataFrame:

        x = X.to_numpy()
        binary_dataset_arr = np.zeros((x.shape[0], len(conditions)), dtype=int)
        conditions_names = []

        for i, condition in enumerate(conditions):
            condition.evaluate_mask(binary_dataset_arr, x, column_index=i)
            conditions_names.append(str(condition))

        binary_dataset = pd.DataFrame(binary_dataset_arr, columns=conditions_names)
            
        return binary_dataset

    def _get_top_conditions(self,binary_dataset, percent):
        if self.type == "classification":
            importances_TOP = []
            for j in range(0, self.condition_importances_.shape[1] + 0, 2):
                class_importances = (
                    self.condition_importances_.iloc[:, j]
                    .replace("-", np.nan)
                    .dropna()
                )
                class_importances_TOP_number = np.round(
                    (percent) * class_importances.shape[0]
                )

                if class_importances_TOP_number == 0:
                    class_importances_TOP_number = 1

                class_importances_TOP = class_importances.loc[
                    0 : class_importances_TOP_number - 1
                ]
                importances_TOP.extend(list(class_importances_TOP))

            importances_TOP_list = list(set(importances_TOP))
        else:

            importances_TOP_number = np.round(
                (percent) * self.condition_importances_.shape[0]
            )
            if importances_TOP_number == 0:
                importances_TOP_number = 1

            importances_TOP = self.condition_importances_.loc[
                0 : importances_TOP_number - 1
            ]
            importances_TOP_list = importances_TOP["conditions"].to_list()
            
        return binary_dataset[importances_TOP_list]


class RuleExplainer(BaseExplainer):
    def __init__(
        self, model, X: pd.DataFrame, y: Labels, type: str = "classification"
    ) -> None:
        """RuleExplainer

        Parameters
        ----------
        model : Model = Union[RuleClassifier, RuleRegressor, SurvivalRules, CN2UnorderedClassifier, CN2SDUnorderedClassifier, DecisionTreeClassifier, DecisionTreeRegressor, SurvivalTree, List[str]]
            Model to be analyzed. RuleXai supports the following Rule models:
             - RuleKit(https://adaa-polsl.github.io/RuleKit-python/): RuleClassifier, RuleRegressor, SurvivalRules
             - Orange (https://orangedatamining.com/): CN2UnorderedClassifier, CN2SDUnorderedClassifier
            It can also extract rules from decision trees:
             - scikit-learn (https://scikit-learn.org/stable/): DecisionTreeClassifier, DecisionTreeRegressor
             - scikit-survival (https://scikit-survival.readthedocs.io/en/stable/): SurvivalTree
            Or you can provide a list of rules as:
             - classification:
                IF attribute1 = (-inf, value) AND  ...  AND attribute2 = <value1, value2) THEN label_atrribute = {class_name}
             - regression:
                IF attribute1 = (-inf, value) AND  ...  AND attribute2 = <value1, value2) THEN target_attribute = {value}
             - survival:
                IF attribute1 = (-inf, value) AND  ...  AND attribute2 = <value1, value2) THEN survival_status_attribute = {survival_status}
        X : pd.DataFrame
            The training dataset used during provided model training
        y : Union[pd.DataFrame, pd.Series]
            The target values (class labels, real number, survival status) used during provided model training
        type : str = None
            The type of problem that the provided model solves. You can choose between:
                - "classification"
                - "regression"
                - "survival"
                default: "classification"
        Attributes
        ----------
        condition_importances_ : pd.DataFrame
            Computed conditions importances
        feature_importances_ : pd.DataFrame
            Feature importances computed base on conditions importances
        """

        if isinstance(y, pd.DataFrame):
            label_name = y.columns[0]
        else:
            label_name = y.name

        if type == "regression":
            model = RegressionModel(model = model, feature_names= X.columns, label_name=label_name)
        elif type == "survival":
            model = SurvivalModel(model=model, feature_names = X.columns, survival_status_name=label_name)
        else:
                                        
            model = ClassificationModel(model = model, feature_names = X.columns, class_names = np.unique(y), label_name = label_name)

        super().__init__(model, X, y, type)

class Explainer(BaseExplainer):
    def __init__(
        self, X: pd.DataFrame, model_predictions: Labels, type: str = "classification"
    ) -> None:
        """Explainer

        Parameters
        ----------
        X : pd.DataFrame
            The training dataset used during provided model training
        model_predictions : Union[pd.DataFrame, pd.Series]
            The training dataset used during provided model training
        type : str
            The type of problem that the provided model solves. You can choose between:
                - "classification"
                - "regression"
            default: "classification"
        Attributes
        ----------
        condition_importances_ : pd.DataFrame
            Computed conditions importances on given dataset
        feature_importances_ : pd.DataFrame
            Feature importances computed base on conditions importances
        """

        
        if (not isinstance(model_predictions, pd.DataFrame)) and (not isinstance(model_predictions, pd.Series)) : 
              model_predictions =  pd.DataFrame(model_predictions, columns=["class"])                
        self._bb_model = BlackBoxModel(X, model_predictions, type)
        super().__init__(None, X, model_predictions, type)

    def explain(self, measure: str = "C2", basic_conditions: bool = False, X_org = None):
        """Compute conditions importances. The importances of a conditions are computed base on: \n
        Marek Sikora: Redefinition of Decision Rules Based on the Importance of Elementary Conditions Evaluation. Fundam. Informaticae 123(2): 171-197 (2013) \n
        https://dblp.org/rec/journals/fuin/Sikora13.html

        Parameters
        ----------
        measure: str
            Specifies the measure that is used to evaluate the quality of the rules. Possible measures for classification and regression problem are: C2, Lift, Correlation. Default: C2. It is not possible to select a measure for the survival problem, the LogRank test is used by default 
        basic_conditions : bool
            Specifies whether to evaluate the conditions contained in the input rules, or to break the conditions in the rules into base conditions so that individual conditions do not overlap
        X_org:
            The dataset on which the rule-based model should be built. It can be the set on which the black-box model was learned or this set before preprocessing (imputation of missing values, dummification, scaling), because such a set can be handled by the rule model 
        Returns
        -------
        self: Explainer
            Fitted explainer with calculated conditions

        """
        self.if_basic_conditions = basic_conditions

        if X_org is None:
            self.model = self._bb_model.get_rules_model(self.X)
            self.condition_importances_ = self._determine_condition_importances(measure)
        else:
            self.model = self._bb_model.get_rules_model(X_org)
            self.condition_importances_ = self._determine_condition_importances(measure,X_org)
 
        self.feature_importances_ = self._determine_feature_importances(self.condition_importances_)

        return self


from rulekit._operator import BaseOperator
import numpy as np
from .rule import Rule, CompoundCondition, ElementaryCondition
from rulekit import RuleKit
from rulekit.classification import RuleClassifier
from rulekit.regression import RuleRegressor
from rulekit.params import Measures
import pandas as pd
from typing import Dict
import importlib

class Model:
    def __init__(
        self, model, feature_names=None, class_names=None, label_name=None
    ) -> None:
        self.model = model
        self.rules = self.get_rules(self.model, feature_names, class_names, label_name)
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(feature_names))}


    def _map_rules_from_RuleKit(self, rules):
        preprocessed_Rules = []
        for rule in rules:
                                      
            preprocessed_Rules.append(self._map_rule_from_RuleKit(rule.__str__()))
        return preprocessed_Rules

    def _map_rules_from_sklearn(self, rules):
        preprocessed_Rules = []
        for rule in rules:
                                      
            preprocessed_Rules.append(self._map_rule_from_sklearn(rule.__str__()))
        return preprocessed_Rules

    def _map_rule_from_sklearn(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")
        compoundCondition = CompoundCondition()

        for condition in conditions:
            if "<=" in condition:
                attribute, value = condition.split(" <= ")
                elementaryCondition = ElementaryCondition(
                    attribute, ElementaryCondition.minus_inf, float(value), False, True, column_index=self.column_indexes[attribute]
                                           
                )
            elif "<" in condition:
                attribute, value = condition.split(" < ")
                elementaryCondition = ElementaryCondition(
                    attribute, ElementaryCondition.minus_inf, float(value), False, False, column_index=self.column_indexes[attribute]
                                            
                )
            elif ">=" in condition:
                attribute, value = condition.split(" >= ")
                elementaryCondition = ElementaryCondition(
                                     
                    attribute, float(value), ElementaryCondition.inf, True, False, column_index=self.column_indexes[attribute]
                )
            elif ">" in condition:
                attribute, value = condition.split(" > ")
                elementaryCondition = ElementaryCondition(
                                     
                    attribute, float(value), ElementaryCondition.inf, False, False, column_index=self.column_indexes[attribute]
                )
            else:
                attribute, value = condition.split(" = ")
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att, consequence_value = consequence.split(" = ")
        consequence_value = consequence_value[1:-1]
        consequence = ElementaryCondition(consequence_att, consequence_value, column_index=self.column_indexes[attribute])

        return self._preprocessRule(Rule(compoundCondition, consequence))

    def _map_rules_from_list(self, rules):
        preprocessed_Rules = []
        for rule in rules:
            preprocessed_Rules.append(self._map_string_rule(rule))
        return preprocessed_Rules

    def _map_string_rule(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")
        compoundCondition = CompoundCondition()

        for condition in conditions:
            attribute, valueset = condition.split(" = ")
            if "," in valueset:
                left, right = valueset.split(",")
                leftClosed = False if left[0] == "(" else True
                rightClosed = False if right[-1:] == ")" else True
                left = left[1:]
                right = right[1:-1]
                elementaryCondition = ElementaryCondition(
                                                  
                    attribute, float(left), float(right), leftClosed, rightClosed, column_index=self.column_indexes[attribute]
                )
            else:
                value = valueset[1:-1]
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att, consequence_value = consequence.split(" = ")
        consequence_value = consequence_value[1:-1]
        consequence = ElementaryCondition(consequence_att, consequence_value, column_index=self.column_indexes[attribute])

        return self._preprocessRule(Rule(compoundCondition, consequence))

    def _preprocessRule(self, rule):
        conditions_for_attributes = {}

        for condition in rule.premise.subconditions:
            attr = condition.attribute

            if attr in conditions_for_attributes.keys():
                old_condition = conditions_for_attributes[attr]
                condition = old_condition.get_intersection(condition)
            conditions_for_attributes[attr] = condition

        premise = CompoundCondition()
        premise.add_subconditions(conditions_for_attributes.values())
        rule.premise = premise

        return rule

    def _get_rules_from_DT(self, tree, feature_names, class_names, label_name):
        tree_ = tree.tree_
        feature_name = [
            feature_names[i] if i != self._tree.TREE_UNDEFINED else "undefined!"
            for i in tree_.feature
        ]
        paths = []
        path = []

        def recurse(node, path, paths):
            if tree_.feature[node] != self._tree.TREE_UNDEFINED:
                name = feature_name[node]
                threshold = tree_.threshold[node]
                p1, p2 = list(path), list(path)
                p1 += [f"{name} <= {threshold}"]
                recurse(tree_.children_left[node], p1, paths)
                p2 += [f"{name} > {threshold}"]
                recurse(tree_.children_right[node], p2, paths)
            else:
                path += [(tree_.value[node], tree_.n_node_samples[node])]
                paths += [path]

        recurse(0, path, paths)

        # sort by samples count
        samples_count = [p[-1][1] for p in paths]
        ii = list(np.argsort(samples_count))
        paths = [paths[i] for i in reversed(ii)]

        rules = []
        for path in paths:
            rule = "IF "
            for p in path[:-1]:
                if rule != "IF ":
                    rule += " AND "
                rule += str(p)
            rule += " THEN "
            if class_names is None:
                rule += (
                                        
                    f"{label_name} = " + "{" + str(path[-1][0][0][0]) + "}"
                )
            else:
                classes = path[-1][0][0]
                l = np.argmax(classes)
                rule += f"{label_name} = " + "{" + f"{class_names[l]}" + "}"
            rules += [rule]

        return rules


class ClassificationModel(Model):
    def __init__(
        self, model, feature_names=None, class_names=None, label_name=None
    ) -> None:
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(feature_names))}
        super().__init__(
            model,
            feature_names=feature_names,
            class_names=class_names,
            label_name=label_name,
        )

    def get_rules(self, model, feature_names=None, class_names=None, label_name=None):

        if isinstance(model, BaseOperator):
            return self._map_rules_from_RuleKit(model.model.rules)
        if isinstance(model, list):
            return self._map_rules_from_list(model)
        sklearn_tree_module = importlib.import_module("sklearn.tree")
        BaseDecisionTree = getattr(sklearn_tree_module, "BaseDecisionTree")
        if isinstance(model, BaseDecisionTree):
            self._tree = getattr(sklearn_tree_module, "_tree")
            rules = self._get_rules_from_DT(
                self.model, feature_names, class_names, label_name
            )
            return self._map_rules_from_sklearn(rules)
        orange_module = importlib.import_module("Orange.classification.rules")
        _RuleClassifier = getattr(orange_module, "_RuleClassifier")
        if isinstance(model, _RuleClassifier):
            return self._map_rules_from_Orange(model.rule_list)

    def _map_rule_from_RuleKit(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")

        compoundCondition = CompoundCondition()

        for condition in conditions:
            attribute, valueset = condition.split(" = ")
            if "," in valueset:
                left, right = valueset.split(",")
                leftClosed = False if left[0] == "(" else True
                rightClosed = False if right[-1:] == ")" else True
                left = left[1:]
                right = right[1:-1]
                elementaryCondition = ElementaryCondition(
                                                  
                    attribute, float(left), float(right), leftClosed, rightClosed, column_index=self.column_indexes[attribute]
                )
            else:
                value = valueset[1:-1]
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att, consequence_value = consequence.split(" = ")
        consequence_value = consequence_value[1:-1]
        consequence = ElementaryCondition(consequence_att, consequence_value, column_index=self.column_indexes[attribute])

        return Rule(compoundCondition, consequence)

    def _map_rules_from_Orange(self, rules):
        preprocessed_Rules = []
        for rule in rules:
            if "TRUE" not in rule.__str__():
                                          
                preprocessed_Rules.append(self._map_rule_from_Orange(rule.__str__()))
        return preprocessed_Rules

    def _map_rule_from_Orange(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")
        compoundCondition = CompoundCondition()

        for condition in conditions:
            if "<=" in condition:
                attribute, value = condition.split("<=")
                elementaryCondition = ElementaryCondition(
                    attribute, ElementaryCondition.minus_inf, float(value), False, True, column_index=self.column_indexes[attribute]
                                           
                )
            elif "<" in condition:
                attribute, value = condition.split("<")
                elementaryCondition = ElementaryCondition(
                    attribute, ElementaryCondition.minus_inf, float(value), False, False, column_index=self.column_indexes[attribute]
                                            
                )
            elif ">=" in condition:
                attribute, value = condition.split(">=")
                elementaryCondition = ElementaryCondition(
                                     
                    attribute, float(value), ElementaryCondition.inf, True, False, column_index=self.column_indexes[attribute]
                )
            elif ">" in condition:
                attribute, value = condition.split(">")
                elementaryCondition = ElementaryCondition(
                                     
                    attribute, float(value), ElementaryCondition.inf, False, False, column_index=self.column_indexes[attribute]
                )
            elif "=" in condition:
                attribute, value = condition.split("=")
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          
            else:
                elementaryCondition = ElementaryCondition("all", "TRUE")

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att, consequence_value = consequence.split("=")
        consequence_value = consequence_value[:-1]
        consequence = ElementaryCondition(consequence_att, consequence_value, column_index=self.column_indexes[attribute])

        return self._preprocessRule(Rule(compoundCondition, consequence))


class RegressionModel(Model):
    def __init__(self, model, feature_names=None, label_name=None) -> None:
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(feature_names))}
        super().__init__(
            model, feature_names=feature_names, class_names=None, label_name=label_name
        )

    def get_rules(self, model, feature_names=None, class_names=None, label_name=None):
        if isinstance(model, BaseOperator):
            return self._map_rules_from_RuleKit(model.model.rules)
        if isinstance(model, list):
            return self._map_rules_from_list(model)
        sklearn_tree_module = importlib.import_module("sklearn.tree")
        BaseDecisionTree = getattr(sklearn_tree_module, "BaseDecisionTree")
        if isinstance(model, BaseDecisionTree):
            self._tree = getattr(sklearn_tree_module, "_tree")
            rules = self._get_rules_from_DT(
                self.model, feature_names, class_names, label_name
            )
            return self._map_rules_from_sklearn(rules)


    def _map_rule_from_RuleKit(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")

        compoundCondition = CompoundCondition()
                                                        
        for condition in conditions:
            attribute, valueset = condition.split(" = ")
            if "," in valueset:
                left, right = valueset.split(",")
                leftClosed = False if left[0] == "(" else True
                rightClosed = False if right[-1:] == ")" else True
                left = left[1:]
                right = right[1:-1]
                elementaryCondition = ElementaryCondition(
                                                  
                    attribute, float(left), float(right), leftClosed, rightClosed, column_index=self.column_indexes[attribute]
                )
            else:
                value = valueset[1:-1]
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att, consequence_value = consequence.split(" = ")
        consequence_val, consequence_range = consequence_value.split(" ")
        consequence_val = consequence_val[1:-1]
                                          
        consequence = ElementaryCondition(consequence_att, float(consequence_val), column_index=self.column_indexes[attribute])

        return Rule(compoundCondition, consequence)

    def _map_rule_from_sklearn(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")
        compoundCondition = CompoundCondition()

        for condition in conditions:
            if "<=" in condition:
                attribute, value = condition.split(" <= ")
                elementaryCondition = ElementaryCondition(
                    attribute, ElementaryCondition.minus_inf, float(value), False, True, column_index=self.column_indexes[attribute]
                                           
                )
            elif "<" in condition:
                attribute, value = condition.split(" < ")
                elementaryCondition = ElementaryCondition(
                    attribute, ElementaryCondition.minus_inf, float(value), False, False, column_index=self.column_indexes[attribute]
                                            
                )
            elif ">=" in condition:
                attribute, value = condition.split(" >= ")
                elementaryCondition = ElementaryCondition(
                                     
                    attribute, float(value), ElementaryCondition.inf, True, False, column_index=self.column_indexes[attribute]
                )
            elif ">" in condition:
                attribute, value = condition.split(" > ")
                elementaryCondition = ElementaryCondition(
                                     
                    attribute, float(value), ElementaryCondition.inf, False, False, column_index=self.column_indexes[attribute]
                )
            else:
                attribute, value = condition.split(" = ")
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att, consequence_value = consequence.split(" = ")
        consequence_value = consequence_value[1:-1]
                                          
        consequence = ElementaryCondition(consequence_att, float(consequence_value), column_index=self.column_indexes[attribute])

        return self._preprocessRule(Rule(compoundCondition, consequence))

    def _map_string_rule(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")
        compoundCondition = CompoundCondition()

        for condition in conditions:
            attribute, valueset = condition.split(" = ")
            if "," in valueset:
                left, right = valueset.split(",")
                leftClosed = False if left[0] == "(" else True
                rightClosed = False if right[-1:] == ")" else True
                left = left[1:]
                right = right[1:-1]
                elementaryCondition = ElementaryCondition(
                                                  
                    attribute, float(left), float(right), leftClosed, rightClosed, column_index=self.column_indexes[attribute]
                )
            else:
                value = valueset[1:-1]
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att, consequence_value = consequence.split(" = ")
        consequence_value = consequence_value[1:-1]
                                          
        consequence = ElementaryCondition(consequence_att, float(consequence_value), column_index=self.column_indexes[attribute])

        return self._preprocessRule(Rule(compoundCondition, consequence))


class SurvivalModel(Model):
    def __init__(self, model, feature_names=None, survival_status_name=None) -> None:
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(feature_names))}
        super().__init__(
            model,
            feature_names=feature_names,
            class_names=None,
            label_name=survival_status_name,
        )

    def get_rules(self, model, feature_names=None, class_names=None, label_name=None):
        if isinstance(model, BaseOperator):
            return self._map_rules_from_RuleKit(model.model.rules)
        if isinstance(model, list):
            return self._map_rules_from_list(model)
        sksurv_tree_module = importlib.import_module("sksurv.tree")
        SurvivalTree = getattr(sksurv_tree_module, "SurvivalTree")
        if isinstance(model, SurvivalTree):
            sklearn_tree_module = importlib.import_module("sklearn.tree")
            self._tree = getattr(sklearn_tree_module, "_tree")
            rules = self._get_rules_from_DT(
                self.model, feature_names, class_names, label_name
            )
            return self._map_rules_from_sklearn(rules)


    def _map_rule_from_RuleKit(self, rule):
        rule = rule[3:]
        premise, consequence = rule.split(" THEN ")
        conditions = premise.split(" AND ")
        compoundCondition = CompoundCondition()

        for condition in conditions:
            attribute, valueset = condition.split(" = ")
            if "," in valueset:
                left, right = valueset.split(",")
                leftClosed = False if left[0] == "(" else True
                rightClosed = False if right[-1:] == ")" else True
                left = left[1:]
                right = right[1:-1]
                elementaryCondition = ElementaryCondition(
                                                  
                    attribute, float(left), float(right), leftClosed, rightClosed, column_index=self.column_indexes[attribute]
                )
            else:
                value = valueset[1:-1]
                elementaryCondition = ElementaryCondition(attribute, str(value), column_index=self.column_indexes[attribute])
                                          

            compoundCondition.add_subcondition(elementaryCondition)

        consequence_att = "survival_curve"
        consequence_val = ""
        consequence = ElementaryCondition(consequence_att, consequence_val, column_index=0)

        return Rule(compoundCondition, consequence)


class BlackBoxModel:
    def __init__(self, X_model, model_predictions, type) -> None:
        RuleKit.init()

        if type == "regression":
            rulekit_model = RuleRegressor(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
        )
        else:
            rulekit_model = RuleClassifier(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
        )
        self.type = type
        self.y = model_predictions
        self.rulekit_model = rulekit_model


    def get_rules_model(self, X_org):
        
        if isinstance(self.y, pd.DataFrame):
            label_name = self.y.columns[0]
            y = self.y.to_numpy().reshape(self.y.size)
        else:
            label_name = self.y.name
            y=self.y

        for column in X_org.select_dtypes('object').columns.tolist():
            X_org[column] = X_org[column].replace({np.nan: None})
        self.rulekit_model.fit(X_org,y)

        if type == "regression":
            model = RegressionModel(self.rulekit_model,label_name)
        else:          
            model = ClassificationModel(self.rulekit_model, X_org.columns, np.unique(self.y), label_name)

        return model






        






import pandas as pd



class Reduct:

    def __init__(self) -> None:
        pass

    def calculate_POS(self, reduct: pd.DataFrame, y: pd.DataFrame):
        POS = []
        
        reduct["if_duplicated"] = reduct.duplicated(keep=False)
        positive_region = [reduct.iloc[i,:].to_numpy() for i in range(reduct.shape[0]) if reduct.if_duplicated.iloc[i] == False]

        reduct["label"] = y
        df_duplicated = reduct[reduct.if_duplicated == True]
        df_duplicated = df_duplicated.drop(["if_duplicated"], axis=1)

        df_duplicated.sort_values(by = df_duplicated.columns.to_list(), inplace = True)
        df_duplicated.reset_index(inplace= True, drop = True)

        df_labels = df_duplicated["label"]
        df_duplicated = df_duplicated.drop(["label"], axis = 1)

        start_equivalence = 0
        for i in range(1, df_duplicated.shape[0]):
            if (not df_duplicated.iloc[i-1,:].equals(df_duplicated.iloc[i,:])) or i == (df_duplicated.shape[0] - 1):
                
                if i == (df_duplicated.shape[0] - 1):
                    end_equivalence = i + 1
                else: 
                    end_equivalence = i
                
                Xj = df_duplicated.iloc[start_equivalence:end_equivalence, :]
                Xj = Xj.assign(label = df_labels.iloc[start_equivalence:end_equivalence])

                if len(Xj["label"].unique()) == 1:
                    positive_region.extend([Xj.iloc[i,:].to_numpy() for i in range(Xj.shape[0])])
                
                start_equivalence = end_equivalence

        return positive_region

    
    def eliminate_irrelevant_attributes(self, R: pd.DataFrame, y: pd.DataFrame):

        POS_R = self.calculate_POS(R.copy(), y)
        m = len(POS_R)
        A = R.columns.to_list()
        for a in A:
            POS_R_a = self.calculate_POS(R.drop([a], axis = 1), y)
            m_a = len(POS_R_a)
            if m_a == m:
                R = R.drop([a], axis = 1)
        
        return R

    def eliminate_irrelevant_attributes_based_on_user_POS(self,R: pd.DataFrame, y: pd.DataFrame, user_POS: float):

        POS_R = self.calculate_POS(R.copy(), y)
        all_objects_number = len(R)
        m = len(POS_R)/all_objects_number
        
        if user_POS > m:
            print("Warning! The set POS is greater than the reduct POS. The reduct will not be limited")
            return R 

        A = R.columns.to_list()
        for a in A:
            POS_R_a = self.calculate_POS(R.drop([a], axis = 1), y)
            m_a = len(POS_R_a)/all_objects_number
            if m_a >= user_POS:
                R = R.drop([a], axis = 1)
        return R

    # Jonhson's algorithm
    def semi_minimal_reduct(self, x: pd.DataFrame):

        repeate = True
        R = pd.DataFrame()
        L = [x]
        A = x.columns.to_list()

        while(repeate):

            W_U_R = dict()
            for a in A:            
                W_U_R[a] = 0
                for Xi in L:
                    cardinalities = Xi[a].value_counts()
                    W_X_i = (pow(cardinalities.sum(), 2) -
                            sum([pow(xi, 2) for xi in cardinalities]))/2
                    W_U_R[a] += W_X_i

            # attribute with the largest W_U_R(a)
            a = max(W_U_R, key=lambda key: W_U_R[key])

            A.remove(a)
            R[a] = x[a].copy()

            tmp_L = []
            for Xi in L:
                tmp_L.extend([Xi[Xi[a] == value] for value in Xi[a].unique()])
            L = tmp_L

            if W_U_R[a] == 0 or len(A) == 0:
                repeate = False

        return R


    def get_reduct(self, x: pd.DataFrame, y: pd.DataFrame = None, POS: float = None):
        
        R = self.semi_minimal_reduct(x)
        if y is None:
            return R 

        if POS is None:
            R = self.eliminate_irrelevant_attributes(R, y)
        else:
            R = self.eliminate_irrelevant_attributes_based_on_user_POS(R,y,POS)
        return R

  


import unittest
import rulekit

from rulekit.main import RuleKit
import pandas as pd
from rulekit import RuleKit
from rulekit.classification import RuleClassifier
from rulekit.regression import RuleRegressor
from rulekit.survival import SurvivalRules
from rulekit.params import Measures
from scipy.io import arff
import numpy as np
import os 
from pathlib import Path
from rulexai.explainer import RuleExplainer

class TestConditionsImportances(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        RuleKit.init()

    
    def test_classification(self):
        classification_resources = "resources/classification/"
        dataset_path = classification_resources + "iris.arff"
        results_path = classification_resources + "results.csv" 
        results_split_path = classification_resources + "results_split.csv" 
        dataset_path = os.path.join(Path(__file__).resolve().parent, dataset_path)
        results_path = os.path.join(Path(__file__).resolve().parent, results_path)
        results_split_path = os.path.join(Path(__file__).resolve().parent, results_split_path)

        train_df = pd.DataFrame(arff.loadarff(dataset_path)[0])

        # code to change encoding of the file
        tmp_df = train_df.select_dtypes([object])
        tmp_df = tmp_df.stack().str.decode("utf-8").unstack()
        for col in tmp_df:
            train_df[col] = tmp_df[col].replace({"?": None})

        x = train_df.drop(["class"], axis=1)
        y = train_df["class"]

        # RuleKit
        clf = RuleClassifier(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
        )
        clf.fit(x, y)


        # RuleXai
        explainer = RuleExplainer(clf, x, y, type="classification")
        explainer.explain()
        conditions_importances = explainer.condition_importances_
        conditions_importances.replace("-", np.nan, inplace = True)
        conditions_importances = conditions_importances.round(4)

        importances_ground_truth = pd.read_csv(results_path, sep = ";")
        importances_ground_truth = importances_ground_truth.round(4)

        # with splitting conditions to basic
        explainer.explain(basic_conditions=True)
        basic_conditions_importances= explainer.condition_importances_
        basic_conditions_importances.replace("-", np.nan, inplace = True)
        basic_conditions_importances = basic_conditions_importances.round(4)

        basic_importances_ground_truth = pd.read_csv(results_split_path, sep = ";")
        basic_importances_ground_truth = basic_importances_ground_truth.round(4)

        assert conditions_importances.equals(importances_ground_truth)
        assert basic_conditions_importances.equals(basic_importances_ground_truth)


    def test_regression(self):

        regression_resources = "resources/regression/"
        dataset_path = regression_resources + "diabetes.arff"
        results_path = regression_resources + "results.csv" 
        results_split_path = regression_resources + "results_split.csv" 
        dataset_path = os.path.join(Path(__file__).resolve().parent, dataset_path)
        results_path = os.path.join(Path(__file__).resolve().parent, results_path)
        results_split_path = os.path.join(Path(__file__).resolve().parent, results_split_path)

        train_df = pd.DataFrame(arff.loadarff(dataset_path)[0])

        # code to change encoding of the file
        tmp_df = train_df.select_dtypes([object])
        if (tmp_df.size != 0):  
            tmp_df = tmp_df.stack().str.decode("utf-8").unstack()
            for col in tmp_df:
                train_df[col] = tmp_df[col].replace({"?": None})

        x = train_df.drop(["class"], axis=1)
        y = train_df["class"]

        df = train_df

        # RuleKit
        reg = RuleRegressor(
            induction_measure=Measures.C2,
            pruning_measure=Measures.C2,
            voting_measure=Measures.C2,
            mean_based_regression = False
        )
        reg.fit(x, y)


        # RuleXai
        explainer = RuleExplainer(reg, x, y, type="regression")
        explainer.explain()
        conditions_importances = explainer.condition_importances_
        conditions_importances = conditions_importances.round(4)

        importances_ground_truth = pd.read_csv(results_path, sep = ";")
        importances_ground_truth = importances_ground_truth.round(4)

        # with splitting conditions to basic
        explainer.explain(basic_conditions=True)
        basic_conditions_importances= explainer.condition_importances_
        basic_conditions_importances = basic_conditions_importances.round(4)

        basic_importances_ground_truth = pd.read_csv(results_split_path, sep = ";")
        basic_importances_ground_truth = basic_importances_ground_truth.round(4)

        assert conditions_importances.equals(importances_ground_truth)
        assert basic_conditions_importances.equals(basic_importances_ground_truth)


 

if __name__ == '__main__':
    unittest.main()


from .rule import (
    CompoundCondition,
    CompoundConditionWithCombiningOperators,
    ClassificationRule,
    ElementaryCondition,
    RegressionRule,
    Rule,
    SurvivalRule,
)
import pandas as pd
import numpy as np
from typing import List
from operator import attrgetter
import math
from typing import Dict
import importlib

class ConditionImportance:
    def __init__(self, rules, data, labels, if_split, measure = None) -> None:
        self.data = data
        self.labels = labels
        self.dataset = pd.concat([self.data, self.labels], axis=1)
        self.if_split = if_split
        self.measure = measure
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(data.columns))}
        if if_split:
            self.rules = self.split_conditions_in_rules(rules)
        else:
            self.rules = rules

    def _get_conditions_with_rules(self, rules):
        conditions_with_rules = dict()

        for rule in rules:
            rule_conditions = rule.premise.get_subconditions()

            for condition in rule_conditions:
                if condition in conditions_with_rules.keys():
                    conditions_rules = conditions_with_rules[condition]
                else:
                    conditions_rules = []
                conditions_rules.append(rule)
                conditions_with_rules[condition] = conditions_rules

        return conditions_with_rules

    def _calculate_conditions_qualities(self, conditions_with_rules):
        conditions_qualities = []
        for condition in conditions_with_rules.keys():
            sum = 0
            for rule in conditions_with_rules[condition]:
                sum += self._calculate_index_simplified(condition, rule)
            conditions_qualities.append(ConditionEvaluation(condition, sum))

        return conditions_qualities


    def _calculateMeasure(self, rule):
        p, n, P, N = rule.covers(self.dataset)

        if self.measure == "Correlation":
            if (P - p + N - n == 0):
                return  0
            else: 
                return (p * N - P * n) / math.sqrt(P * N * (p + n) * (P - p + N - n))

        elif self.measure == "Lift":
            if (p == 0 and n == 0) or P == 0:
                return 0
            else:
                return p * (P + N) / ((p + n) * P)

        else: # C2
            if (p == 0 and n == 0) or P == 0 or N == 0:
                return 0
            else:
                return (((P + N) * p / (p + n) - P) / N) * ((1 + p / P) / 2)  # C2

    def _condition_importances_to_DataFrame(self, condition_importances):
        importances_df = pd.DataFrame()
        importances_df["conditions"] = pd.Series(
            [str(cnd) for cnd in condition_importances.keys()]
        )
        importances_df["importances"] = pd.Series(condition_importances.values())

        return importances_df

    def split_conditions_in_rules(self, rules):
        conditions_with_rules = self._get_conditions_with_rules(rules)
        conditions = self._split_conditions_into_basic(conditions_with_rules.keys())

        return self._get_rules_with_splitted_conditions(conditions, rules)

    def _split_conditions_into_basic(self, conditions: List[ElementaryCondition]):
        splittedConditions = []
        conditions_for_attributes = dict()
        for condition in conditions:
            if condition.right is None: #attribute is Nominal
                splittedConditions.append(condition)
            else:
                attr = condition.attribute
                if attr in conditions_for_attributes:
                    attribute_conditions = conditions_for_attributes[attr]
                else: 
                    attribute_conditions = []
                attribute_conditions.append(condition)
                conditions_for_attributes[attr] = attribute_conditions

        for conditions_for_attribute in conditions_for_attributes.values():
            if len(conditions_for_attribute) > 1:
                basic_conditions = self._get_basic_conditions_for_attribute(conditions_for_attribute)
                splittedConditions.extend(basic_conditions)
            else:
                splittedConditions.extend(conditions_for_attribute)

        return splittedConditions
    
    def _get_basic_conditions_for_attribute(self, conditions: List[ElementaryCondition]):
        basic_conditions = []
        points = []
        id = 0 
        attribute = conditions[0].attribute
        for condition in conditions:
            point = Point(id, condition.left, "Left", condition.leftClosed)
            points.append(point)
            point = Point(id, condition.right, "Right", condition.rightClosed)
            points.append(point)
            id += 1

        min_point_first = min(points, key = attrgetter('value'))
        all_points_len = len(points)
        points = [point for point in points if point.value != min_point_first.value]
        number_of_firts_mins = all_points_len - len(points)

        while(len(points) > 0):
            min_point_second = min(points, key = attrgetter('value'))
            points_with_second_min_len = len(points)
            points = [point for point in points if point.value != min_point_second.value]
            number_of_second_mins = points_with_second_min_len - len(points)

            if(min_point_first.value == float('-inf')):
                leftClosed = False
            else:
                leftClosed = True

            condition = ElementaryCondition(attribute = attribute, left = min_point_first.value, right= min_point_second.value, leftClosed = leftClosed, rightClosed= False, column_index=self.column_indexes[attribute])
            basic_conditions.append(condition)

            if (min_point_first.condition_id == min_point_second.condition_id) and number_of_firts_mins == 1 and number_of_second_mins == 1 and len(points) > 1:
                min_point_first = min(points, key = attrgetter('value'))
                points_with_first_min_len = len(points)
                points = [point for point in points if point.value != min_point_first.value]
                number_of_firts_mins = points_with_first_min_len - len(points)
            else:
                min_point_first = min_point_second
                min_point_first.is_closed = not min_point_second.is_closed

        return basic_conditions

    
    def _get_rules_with_splitted_conditions(self, basic_conditions, rules: List[ClassificationRule]):
        rules_with_basic_conditions = []
        for rule in rules:
            compoundCondition = self._create_compound_condition_for_rule(basic_conditions, rule)
            rules_with_basic_conditions.append(Rule(compoundCondition, rule.consequence))
        return rules_with_basic_conditions

    def _create_compound_condition_for_rule(self,basic_conditions, rule: Rule):
        subconditions = rule.premise.get_subconditions()
        compoundCondition = CompoundConditionWithCombiningOperators()
        for condition_basic in basic_conditions:
            i = 0
            condition_added = False
            while(i<len(subconditions) and not condition_added):
                condition_from_rule = subconditions[i]
                if self._check_if_condition_contains_basic_condition(condition_from_rule, condition_basic):
                    compoundCondition.add_subcondition(condition_basic)
                    condition_added = True
                i += 1
        return compoundCondition

    def _check_if_condition_contains_basic_condition(self,condition_from_rule: ElementaryCondition, basic_condition: ElementaryCondition):
        if not condition_from_rule.attribute == basic_condition.attribute:
            return False
        elif condition_from_rule.right == None: #nominal attribute
            if condition_from_rule.left == basic_condition.left:
                return True
            else:
                return False
        else:
            if ((condition_from_rule.left <= basic_condition.left) and (condition_from_rule.right >= basic_condition.right)):
                return True
            else:
                return False
            

class ClassificationConditionImportance(ConditionImportance):
    def __init__(self, rules, data, labels, if_split, measure) -> None:
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(data.columns))}
        super().__init__(rules, data, labels, if_split, measure)

    def condition_importances(self):
        rules_by_class = self._split_rules_by_decision_class(self.rules)

        condition_importances_for_classes = dict()

        for class_name in rules_by_class.keys():
            class_rules = rules_by_class[class_name]
            conditions_with_rules = self._get_conditions_with_rules(class_rules)
            conditions_qualities = self._calculate_conditions_qualities(
                conditions_with_rules
            )
            condition_importances_for_classes[class_name] = conditions_qualities

        conditions_importances = self._calculate_conditions_importances(
            condition_importances_for_classes
        )
        return self._condition_importances_to_DataFrame(conditions_importances)

    def _split_rules_by_decision_class(self, rules):
        rules_by_class = dict()

        for rule in rules:
            class_name = rule.consequence.left
            if class_name in rules_by_class.keys():
                class_rules = rules_by_class[class_name]
            else:
                class_rules = []

            class_rules.append(rule)
            rules_by_class[class_name] = class_rules

        return rules_by_class

    def _calculate_index_simplified(self, condition, rule):
        rule = ClassificationRule(rule.premise, rule.consequence)

        rule_conditions = []
        rule_conditions.extend(rule.premise.get_subconditions())
        number_of_conditions = len(rule_conditions)
        rule_conditions.remove(condition)

        if self.if_split:
            premise_without_evaluated_condition = CompoundConditionWithCombiningOperators()
        else:
            premise_without_evaluated_condition = CompoundCondition()
        
        premise_without_evaluated_condition.add_subconditions(rule_conditions)
        rule_without_evaluated_condition = ClassificationRule(
            premise_without_evaluated_condition, rule.consequence
        )

        factor = 1.0 / number_of_conditions

        if len(rule_conditions) == 0:
            return factor * (
                self._calculateMeasure(rule)
                - self._calculateMeasure(rule_without_evaluated_condition)
            )
        else:
            premise_with_only_evaluated_condition = CompoundCondition()
            premise_with_only_evaluated_condition.add_subcondition(condition)
            rule_with_only_evaluated_condition = ClassificationRule(
                premise_with_only_evaluated_condition, rule.consequence
            )

            return factor * (
                self._calculateMeasure(rule)
                - self._calculateMeasure(rule_without_evaluated_condition)
                + self._calculateMeasure(rule_with_only_evaluated_condition)
            )

    def _calculate_conditions_importances(self, condition_qualities_for_classes):
        conditions_importances = dict()

        for evaluated_class in condition_qualities_for_classes.keys():

            conditions_importances_for_class = dict()
            conditions_for_evaluated_class = condition_qualities_for_classes[
                evaluated_class
            ]

            for condition in conditions_for_evaluated_class:
                sum = condition.quality
                """ 
                for class_name in condition_qualities_for_classes.keys():
                    if class_name != evaluated_class:
                        conditions_from_other_class = condition_qualities_for_classes[
                            class_name
                        ]
                        evaluated_condition_from_other_class = list(
                            filter(
                                lambda cnd: cnd.condition == condition.condition,
                                conditions_from_other_class,
                            )
                        )
                        if len(evaluated_condition_from_other_class) > 0:
                            sum -= evaluated_condition_from_other_class[0].quality
                """
                conditions_importances_for_class[condition.condition] = sum

            conditions_importances[evaluated_class] = dict(
                sorted(
                    conditions_importances_for_class.items(),
                    key=lambda item: item[1],
                    reverse=True,
                )
            )

        return conditions_importances

    def _condition_importances_to_DataFrame(self, condition_importances):
        importances_df = pd.DataFrame()
        for class_name in condition_importances.keys():
            importances_df_tmp = pd.DataFrame()
            importances_df_tmp[class_name + " | conditions_names"] = pd.Series(
                [str(cnd) for cnd in condition_importances[class_name].keys()]
            )
            importances_df_tmp[class_name + " | importances"] = pd.Series(
                condition_importances[class_name].values()
            )
            importances_df = pd.concat(
                [importances_df, importances_df_tmp], ignore_index=False, axis=1
            )
        return importances_df.replace(np.nan, "-")

    def _get_rules_with_splitted_conditions(self, basic_conditions, rules: List[ClassificationRule]):
        rules_with_basic_conditions = []
        for rule in rules:
            compoundCondition = self._create_compound_condition_for_rule(basic_conditions, rule)
            rules_with_basic_conditions.append(ClassificationRule(compoundCondition, rule.consequence))
        return rules_with_basic_conditions



class RegressionConditionImportance(ConditionImportance):
    def __init__(self, rules, data, labels, if_split, measure) -> None:
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(data.columns))}
        super().__init__(rules, data, labels, if_split, measure)

    def condition_importances(self):
        conditions_with_rules = self._get_conditions_with_rules(self.rules)
        conditions_qualities = self._calculate_conditions_qualities(
            conditions_with_rules
        )
        conditions_importances = dict()

        for condition_evaluation in conditions_qualities:
            conditions_importances[
                condition_evaluation.condition
            ] = condition_evaluation.quality

        conditions_importances = dict(
            sorted(
                conditions_importances.items(), key=lambda item: item[1], reverse=True
            )
        )

        return self._condition_importances_to_DataFrame(conditions_importances)

    def _calculate_index_simplified(self, condition, rule):
        rule = RegressionRule(rule.premise, rule.consequence)

        rule_conditions = []
        rule_conditions.extend(rule.premise.get_subconditions())
        number_of_conditions = len(rule_conditions)
        rule_conditions.remove(condition)

        if self.if_split:
            premise_without_evaluated_condition = CompoundConditionWithCombiningOperators()
        else:
            premise_without_evaluated_condition = CompoundCondition()
        premise_without_evaluated_condition.add_subconditions(rule_conditions)
        rule_without_evaluated_condition = RegressionRule(
            premise_without_evaluated_condition, rule.consequence
        )

        factor = 1.0 / number_of_conditions

        if len(rule_conditions) == 0:
            return factor * (
                self._calculateMeasure(rule)
                - self._calculateMeasure(rule_without_evaluated_condition)
            )
        else:
            premise_with_only_evaluated_condition = CompoundCondition()
            premise_with_only_evaluated_condition.add_subcondition(condition)
            rule_with_only_evaluated_condition = RegressionRule(
                premise_with_only_evaluated_condition, rule.consequence
            )

            return factor * (
                self._calculateMeasure(rule)
                - self._calculateMeasure(rule_without_evaluated_condition)
                + self._calculateMeasure(rule_with_only_evaluated_condition)
            )

    def _get_rules_with_splitted_conditions(self, basic_conditions, rules: List[RegressionRule]):
        rules_with_basic_conditions = []
        for rule in rules:
            compoundCondition = self._create_compound_condition_for_rule(basic_conditions, rule)
            rules_with_basic_conditions.append(RegressionRule(compoundCondition, rule.consequence))
        return rules_with_basic_conditions

class SurvivalConditionImportance(ConditionImportance):
    def __init__(self, rules, data, labels, if_split, measure) -> None:
        self.column_indexes: Dict[str, int] = {column_name:i for i, column_name in enumerate(list(data.columns))}
        super().__init__(rules, data, labels, if_split, measure)
        lifelines = importlib.import_module("lifelines.statistics")
        self.logrank_test = getattr(lifelines, "logrank_test")
        

    def condition_importances(self):
        conditions_with_rules = self._get_conditions_with_rules(self.rules)
        conditions_qualities = self._calculate_conditions_qualities(
            conditions_with_rules
        )
        conditions_importances = dict()

        for condition_evaluation in conditions_qualities:
            conditions_importances[
                condition_evaluation.condition
            ] = condition_evaluation.quality

        conditions_importances = dict(
            sorted(
                conditions_importances.items(), key=lambda item: item[1], reverse=True
            )
        )

        return self._condition_importances_to_DataFrame(conditions_importances)

    def _calculate_index_simplified(self, condition, rule):
        rule = SurvivalRule(rule.premise, rule.consequence)

        rule_conditions = []
        rule_conditions.extend(rule.premise.get_subconditions())
        number_of_conditions = len(rule_conditions)
        rule_conditions.remove(condition)

        if self.if_split:
            premise_without_evaluated_condition = CompoundConditionWithCombiningOperators()
        else:
            premise_without_evaluated_condition = CompoundCondition()

        premise_without_evaluated_condition.add_subconditions(rule_conditions)
        rule_without_evaluated_condition = SurvivalRule(
            premise_without_evaluated_condition, rule.consequence
        )

        factor = 1.0 / number_of_conditions

        if len(rule_conditions) == 0:
            return factor * (
                self._calculateMeasure(rule)
                - self._calculateMeasure(rule_without_evaluated_condition)
            )
        else:
            premise_with_only_evaluated_condition = CompoundCondition()
            premise_with_only_evaluated_condition.add_subcondition(condition)
            rule_with_only_evaluated_condition = SurvivalRule(
                premise_with_only_evaluated_condition, rule.consequence
            )

            return factor * (
                self._calculateMeasure(rule)
                - self._calculateMeasure(rule_without_evaluated_condition)
                + self._calculateMeasure(rule_with_only_evaluated_condition)
            )

    def _calculateMeasure(self, rule):
        p, n, P, N, covered_indices = rule.covers(self.dataset, return_positives=True
        )
        # covered_indices -> in survival rules all examples are classified as positives
        uncovered_indices = [
            id for id in range(self.data.shape[0]) if id not in covered_indices
        ]

        results = self.logrank_test(
            self.data["survival_time"][
                self.data["survival_time"].index.isin(covered_indices)
            ],
            self.data["survival_time"][
                self.data["survival_time"].index.isin(uncovered_indices)
            ],
            event_observed_A=self.labels[self.labels.index.isin(covered_indices)],
            event_observed_B=self.labels[self.labels.index.isin(uncovered_indices)],
        )

        return results.test_statistic


    def _get_rules_with_splitted_conditions(self, basic_conditions, rules: List[SurvivalRule]):
        rules_with_basic_conditions = []
        for rule in rules:
            compoundCondition = self._create_compound_condition_for_rule(basic_conditions, rule)
            rules_with_basic_conditions.append(SurvivalRule(compoundCondition, rule.consequence))
        return rules_with_basic_conditions



class ConditionEvaluation:
    def __init__(self, condition, quality) -> None:
        self.condition = condition
        self.quality = quality


class Point:
    def __init__(self, condition_id: int, value: float, side: str, is_closed: bool) -> None:
        self.condition_id = condition_id
        self.value = value
        self.side = side
        self.is_closed = is_closed


import setuptools
import os
import io

current_path = os.path.dirname(os.path.realpath(__file__))

with io.open(f"{current_path}/README.md", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="rulexai",
    version="1.1.0",
    author="Dawid Macha",
    author_email="dawid.m.macha@gmail.com",
    description="RuleXAI is a rule-based aproach to explain the output of any machine learning model. It is suitable for classification, regression and survival tasks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/adaa-polsl/RuleXAI",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    include_package_data=True,
    python_requires=">=3.9",
    install_requires=[
        "pandas >= 1.5.0, < 2.3.0",
        "numpy ~= 1.26.4",
        "matplotlib ~= 3.8.3",
        "rulekit ~= 1.7.6",
        "lifelines ~= 0.28.0"
    ],
    test_suite="tests",
)



import enum
import numpy as np
from typing import List
import warnings
warnings.filterwarnings('ignore')

# creating enumerations using class
class LogicalOperator(enum.Enum):
    conjuction = 1
    alternative = 2


class ElementaryCondition:

    inf = float("inf")
    minus_inf = float("-inf")

    def __init__(
        self,
        attribute: str,
        left: float,
        right: float = None,
        leftClosed: bool = None,
        rightClosed: bool = None,
        column_index: int = None,
    ) -> None:
        self.attribute = attribute
        self.left = left
        self.right = right
        self.leftClosed = leftClosed
        # if right is None its means that attribute is Nominal
        self.rightClosed = rightClosed
        self._column_index = column_index

    def covered_mask(self, X: np.ndarray) -> np.ndarray:
        if self.right is not None:
            left_part: np.ndarray = (X[:, self._column_index] >= self.left) if self.leftClosed else (X[:, self._column_index] > self.left)
            right_part: np.ndarray = (X[:, self._column_index] <= self.right) if self.rightClosed else (X[:, self._column_index] < self.right)
            return left_part & right_part
        else:
            return (X[:, self._column_index] == self.left)

    def uncovered_mask(self, X: np.ndarray) -> np.ndarray:
        return np.logical_not(self.covered_mask(X))

    def negative_covered_mask(self, X: np.ndarray, y: np.ndarray, decision) -> np.ndarray:
        return self.covered_mask(X) & (y[:] != decision)

    def positive_covered_mask(self, X: np.ndarray, y: np.ndarray, decision) -> np.ndarray:
        a =  self.covered_mask(X) & (y[:] == decision)
        return a
    def get_intersection(self, other_condition):
        return ElementaryCondition(
            self.attribute,
            np.max([self.left, other_condition.left]),
            np.min([self.right, other_condition.right]),
            other_condition.leftClosed
            if self.left < other_condition.left
            else self.leftClosed,
            other_condition.rightClosed
            if self.right > other_condition.right
            else self.rightClosed,
            column_index=self._column_index
        )

    def evaluate(self, ex):
        value = ex[self.attribute]
        if self.right == None:
            return str(value) == self.left
        else:
            return ((value >= self.left and self.leftClosed) or value > self.left) and (
                (value <= self.right and self.rightClosed) or value < self.right
            )
    def evaluate_mask(self, X_t: np.ndarray, X: np.ndarray, column_index: int = None):
        X_t[:, column_index] = self.covered_mask(X)
        return X_t

    def __str__(self):
        if self.right == None:
            s = "".join(["{", str(self.left), "}"])
        else:
            s = "".join(
                [
                    ("<" if self.leftClosed else "("),
                    ("-inf" if (self.left == self.minus_inf) else str(self.left)),
                    ", ",
                    ("inf" if (self.right == self.inf) else str(self.right)),
                    (">" if self.rightClosed else ")"),
                ]
            )
        return "".join([self.attribute, " = ", s])

    def __eq__(self, other):
        if not isinstance(other, ElementaryCondition):
            return False
        return (
            (self.attribute == other.attribute)
            and (self.left == other.left)
            and (self.right == other.right)
            and (self.leftClosed == other.leftClosed)
            and (self.rightClosed == other.rightClosed)
        )

    def __hash__(self):
        return hash(
            (self.attribute, self.left, self.right,
             self.leftClosed, self.rightClosed)
        )


class CompoundCondition:

    def __init__(self) -> None:
        self.operator = LogicalOperator.conjuction
        self.subconditions = []

    def covered_mask(self, X: np.ndarray) -> np.ndarray:
        if len(self.subconditions) == 0:
            return np.full(X.shape[0], fill_value=False)
        covered_mask = (self.subconditions[0].covered_mask(X))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(self.subconditions)):
                covered_mask &= self.subconditions[i].covered_mask(X)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(self.subconditions)):
                covered_mask |= self.subconditions[i].covered_mask(X)
        return covered_mask

    def uncovered_mask(self, X: np.ndarray) -> np.ndarray:
        if len(self.subconditions) == 0:
            return np.full(X.shape[0], fill_value=True)
        uncovered_mask = (self.subconditions[0].uncovered_mask(X))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(self.subconditions)):
                uncovered_mask &= self.subconditions[i].uncovered_mask(X)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(self.subconditions)):
                uncovered_mask |= self.subconditions[i].uncovered_mask(X)
        return uncovered_mask

    def positive_covered_mask(self, X: np.ndarray, y: np.ndarray, decision) -> np.ndarray:
        if len(self.subconditions) == 0:
            return np.full(X.shape[0], fill_value=False)
        positive_covered_mask = (
            self.subconditions[0].positive_covered_mask(X, y, decision))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(self.subconditions)):
                positive_covered_mask &= self.subconditions[i].positive_covered_mask(
                    X, y, decision)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(self.subconditions)):
                positive_covered_mask |= self.subconditions[i].positive_covered_mask(
                    X, y, decision)
        return positive_covered_mask

    def negative_covered_mask(self, X: np.ndarray, y: np.ndarray, decision) -> np.ndarray:
        if len(self.subconditions) == 0:
            return np.full(X.shape[0], fill_value=False)
        negative_covered_mask = (
            self.subconditions[0].negative_covered_mask(X, y, decision))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(self.subconditions)):
                negative_covered_mask &= self.subconditions[i].negative_covered_mask(
                    X, y, decision)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(self.subconditions)):
                negative_covered_mask |= self.subconditions[i].negative_covered_mask(
                    X, y, decision)
        return negative_covered_mask

    def add_subcondition(self, cnd: ElementaryCondition):
        self.subconditions.append(cnd)

    def add_subconditions(self, cnds: List[ElementaryCondition]):
        self.subconditions.extend(cnds)

    def get_subconditions(self):
        return self.subconditions

    def set_logical_operator(self, operator: LogicalOperator):
        self.operator = operator

    def __str__(self):
        s = ""
        operator = " AND " if (
            self.operator == LogicalOperator.conjuction) else " OR "

        for i in range(len(self.subconditions)):
            s += self.subconditions[i].__str__()
            if i != (len(self.subconditions) - 1):
                s += operator
        return s

    def evaluate(self, ex):
        for condition in self.subconditions:
            partial = condition.evaluate(ex)
            if self.operator == LogicalOperator.conjuction and partial == False:
                return False
            elif self.operator == LogicalOperator.alternative and partial == True:
                return True
        return True if (self.operator == LogicalOperator.conjuction) else False


class CompoundConditionWithCombiningOperators(CompoundCondition):

    def __init__(self) -> None:
        self.operator = LogicalOperator.conjuction
        self.subCompoundConditions = dict()
    
    def covered_mask(self, X: np.ndarray) -> np.ndarray:
        subCompoundConditions_keys = list(self.subCompoundConditions.keys())
        if len(self.subCompoundConditions) == 0:
            return np.full(X.shape[0], fill_value=False)
        covered_mask = (self.subCompoundConditions[subCompoundConditions_keys[0]].covered_mask(X))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(subCompoundConditions_keys)):
                covered_mask &= self.subCompoundConditions[subCompoundConditions_keys[i]].covered_mask(X)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(subCompoundConditions_keys)):
                covered_mask |= self.subCompoundConditions[subCompoundConditions_keys[i]].covered_mask(X)
        return covered_mask

    def uncovered_mask(self, X: np.ndarray) -> np.ndarray:
        subCompoundConditions_keys = list(self.subCompoundConditions.keys())
        if len(self.subCompoundConditions) == 0:
            return np.full(X.shape[0], fill_value=True)
        uncovered_mask = (self.subCompoundConditions[subCompoundConditions_keys[0]].uncovered_mask(X))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(subCompoundConditions_keys)):
                uncovered_mask &= self.subCompoundConditions[subCompoundConditions_keys[i]].uncovered_mask(X)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(subCompoundConditions_keys)):
                uncovered_mask |= self.subCompoundConditions[subCompoundConditions_keys[i]].uncovered_mask(X)
        return uncovered_mask

    def positive_covered_mask(self, X: np.ndarray, y: np.ndarray, decision) -> np.ndarray:
        subCompoundConditions_keys = list(self.subCompoundConditions.keys())
        if len(self.subCompoundConditions) == 0:
            return np.full(X.shape[0], fill_value=False)
        positive_covered_mask = (
            self.subCompoundConditions[subCompoundConditions_keys[0]].positive_covered_mask(X, y, decision))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(subCompoundConditions_keys)):
                positive_covered_mask &= self.subCompoundConditions[subCompoundConditions_keys[i]].positive_covered_mask(
                    X, y, decision)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(subCompoundConditions_keys)):
                positive_covered_mask |= self.subCompoundConditions[subCompoundConditions_keys[i]].positive_covered_mask(
                    X, y, decision)
        return positive_covered_mask

    def negative_covered_mask(self, X: np.ndarray, y: np.ndarray, decision) -> np.ndarray:
        subCompoundConditions_keys = list(self.subCompoundConditions.keys())
        if len(self.subCompoundConditions) == 0:
            return np.full(X.shape[0], fill_value=False)
        negative_covered_mask = (
            self.subCompoundConditions[subCompoundConditions_keys[0]].negative_covered_mask(X, y, decision))
        if self.operator == LogicalOperator.conjuction:
            for i in range(1, len(subCompoundConditions_keys)):
                negative_covered_mask &= self.subCompoundConditions[subCompoundConditions_keys[i]].negative_covered_mask(
                    X, y, decision)
        elif self.operator == LogicalOperator.alternative:
            for i in range(1, len(subCompoundConditions_keys)):
                negative_covered_mask |= self.subCompoundConditions[subCompoundConditions_keys[i]].negative_covered_mask(
                    X, y, decision)
        return negative_covered_mask

    def add_subcondition(self, cnd: ElementaryCondition):
        attr = cnd.attribute
        if attr in self.subCompoundConditions.keys():
            compoundCondition = self.subCompoundConditions[attr]
        else:
            compoundCondition = CompoundCondition()
            compoundCondition.set_logical_operator(LogicalOperator.alternative)
        compoundCondition.add_subcondition(cnd)
        self.subCompoundConditions[attr] = compoundCondition

    def add_subconditions(self, cnds: List[ElementaryCondition]):
        for condition in cnds:
            self.add_subcondition(condition)

    def get_subconditions(self):
        all_conditions_list = []
        for compundCondition in self.subCompoundConditions.values():
            all_conditions_list.extend(compundCondition.get_subconditions())
        return all_conditions_list

    def set_logical_operator(self, operator: LogicalOperator):
        self.operator = operator

    def __str__(self):
        s = ""
        operator = " AND " if (
            self.operator == LogicalOperator.conjuction) else " OR "
        operator_internal = " OR " if (
            self.operator == LogicalOperator.conjuction) else " AND "

        for compound_condition in self.subCompoundConditions.values():
            s += "["
            for condition_base in compound_condition.get_subconditions():
                s += condition_base.__str__() + operator_internal
            s = s[0: len(s) - len(operator_internal)]
            s += "]" + operator

        s = s[0: len(s) - len(operator)]
        return s

    def evaluate(self, ex):
        for compound_condition in self.subCompoundConditions.values():
            partial = compound_condition.evaluate(ex)
            if self.operator == LogicalOperator.conjuction and partial == False:
                return False
            elif self.operator == LogicalOperator.alternative and partial == True:
                return True
        return True if (self.operator == LogicalOperator.conjuction) else False


class Rule:

    def __init__(
        self, premise: CompoundCondition, consequence: ElementaryCondition
    ) -> None:
        self.premise = premise
        self.consequence = consequence

    def __str__(self) -> str:
        return "".join(
            ["IF ", self.premise.__str__(), " THEN ", self.consequence.__str__()]
        )


class ClassificationRule(Rule):

    def __init__(
        self, premise: CompoundCondition, consequence: ElementaryCondition
    ) -> None:
        super().__init__(premise, consequence)
        self.decision = consequence.left

    def covers(self, X_df):
        x = X_df.to_numpy()
        y = x[:, -1].astype(str)
        x = x[:, 0:-1]

        P = y[y == self.decision].shape[0]
        N = y.shape[0] - P
        p = x[self.premise.positive_covered_mask(x, y, self.decision)].shape[0]
        n = x[self.premise.negative_covered_mask(x, y, self.decision)].shape[0]

        return p, n, P, N


class RegressionRule(Rule):

    def __init__(
        self, premise: CompoundCondition, consequence: ElementaryCondition
    ) -> None:
        super().__init__(premise, consequence)

    def covers(self, x):
        P = 0
        N = 0
        p = 0
        n = 0

        sum_y = 0.0
        sum_y2 = 0.0

        label_name = self.consequence.attribute
        x.sort_values(label_name, inplace=True)

        orderedNegatives = []
        negatives = []
        positives = []

        for i in range(x.shape[0]):
            ex = x.iloc[i]
            N += 1

            if self.premise.evaluate(ex):  # if covered
                n += 1
                orderedNegatives.append(i)
                negatives.append(i)

                y = ex[label_name]

                sum_y += y
                sum_y2 += y * y

        if len(negatives) == 0:
            return p, n, P, N

        mean_y = sum_y / n
        stddev_y = np.sqrt(sum_y2 / n - mean_y * mean_y)

        medianId = orderedNegatives[len(orderedNegatives) // 2]

        median_y = x.iloc[medianId][label_name]

        # update positives
        for i in range(x.shape[0]):
            ex = x.iloc[i]

            # if inside epsilon
            if np.abs(ex[label_name] - median_y) <= stddev_y:
                N -= 1
                P += 1

                # if covered
                if self.premise.evaluate(ex):
                    negatives.remove(i)
                    n -= 1
                    positives.append(i)
                    p += 1

        return p, n, P, N


class SurvivalRule(Rule):

    def __init__(
        self, premise: CompoundCondition, consequence: ElementaryCondition
    ) -> None:
        super().__init__(premise, consequence)

    def covers(self, x, return_positives: bool = False):
        P = 0
        N = 0
        p = 0
        n = 0
        if return_positives:
            positives = []

        for i in range(x.shape[0]):
            ex = x.iloc[i]
            P += 1
            if self.premise.evaluate(ex):
                p += 1
                if return_positives:
                    positives.append(i)
        if return_positives:
            return p, n, P, N, positives
        else:
            return p, n, P, N
